from numba.cuda.cudadrv.devicearray import DeviceNDArray
from .kernel import mse_derivate as kernel_mse_derivate

def mse_derivate(predicts: DeviceNDArray, targets: DeviceNDArray, buffer: DeviceNDArray) -> DeviceNDArray:
    """evaluates: error[i] = (2/N) * (predict[i] - target[i])

    Args:
        predicts (DeviceNDArray): [Batch][1][N]
        targets (DeviceNDArray): [Batch][1][N]
        buffer (DeviceNDArray): [Batch][1][N]

    Returns:
        DeviceNDArray: [Batch][1][N]
    """
    kernel_mse_derivate(buffer=buffer, predicts=predicts, targets=targets)
    
    return buffer



import numpy as np
from numba import cuda
from Kinho.lib import cpu, gpu

EPS = 1e-8

def test_stochastic_gradient_descent_gpu():
    BATCH = 256
    DIM = 100
    DIM2 = 200
    
    gradients_host = np.random.randn(BATCH, DIM, DIM2)
    
    gradients_dvc = cuda.to_device(gradients_host)
    buffer = cuda.device_array(shape=(DIM, DIM2), dtype=np.float64)
    
    ret_cpu = cpu.stochastic_gradient_descent(gradients=gradients_host)
    gpu.stochastic_gradient_descent(gradients=gradients_dvc, buffer=buffer)
    ret_gpu = buffer.copy_to_host()
    
    for i in range(DIM):
        for j in range(DIM2):
            assert abs(ret_cpu[i, j] - ret_gpu[i, j]) < EPS



import numpy as np
from numba import cuda
from Kinho.lib import cpu, gpu

EPS = 1e-8

def test_dot_matrix_derivate_gpu():
    BATCH = 512
    DIM1 = 100
    DIM2 = 300
    
    weight_host = np.random.randn(DIM1, DIM2)
    alphas_host = np.random.randn(BATCH, 1, DIM2)
    
    weight_dvc = cuda.to_device(weight_host)
    alphas_dvc = cuda.to_device(alphas_host)
    buffer = cuda.device_array(shape=(BATCH, 1, DIM1), dtype=np.float64)
    
    ret_cpu = cpu.dot_matrix_derivate(const_matrix=weight_host, alphas=alphas_host)
    gpu.dot_matrix_derivate(const_matrix=weight_dvc, alphas=alphas_dvc, buffer=buffer)
    ret_gpu = buffer.copy_to_host()
    
    for i in range(BATCH):
        for j in range(DIM1):
            assert abs(ret_cpu[i, 0, j] - ret_gpu[i, 0, j]) < EPS



from numba.cuda.cudadrv.devicearray import DeviceNDArray
from numba import cuda
from .utils import ceil, grid_config

THREADSPERBLOCK = (8, 128)

@cuda.jit
def perform(buffer: DeviceNDArray, const_A: DeviceNDArray, const_B: DeviceNDArray, eta: DeviceNDArray):
    row, col = cuda.grid(2)
    
    if row >= buffer.shape[0] or \
        col >= buffer.shape[1]:
            return None
    
    buffer[row, col] = const_A[row, col] - eta[0] * const_B[row, col]

def partial_gradient(buffer: DeviceNDArray, const_A: DeviceNDArray, const_B: DeviceNDArray, eta: DeviceNDArray):
    blockspergrid = grid_config(
        (ceil(buffer.shape[0], THREADSPERBLOCK[0]),
         ceil(buffer.shape[1], THREADSPERBLOCK[1]))
    )
    
    perform[(blockspergrid, THREADSPERBLOCK)](buffer, const_A, const_B, eta)
    cuda.synchronize()



from numba.cuda.cudadrv.devicearray import DeviceNDArray
from numba import cuda
from .utils import ceil, grid_config

THREADSPERBLOCK = (256, 4)

@cuda.jit
def perform(buffer: DeviceNDArray):
    batch_id, idx = cuda.grid(2)
    
    if batch_id >= buffer.shape[0] or \
        idx >= buffer.shape[1]:
            return None
    
    buffer[batch_id, idx] = float(0.0)

def memset_dim_1(buffer: DeviceNDArray):
    blockspergrid = grid_config(
        (ceil(buffer.shape[0], THREADSPERBLOCK[0]),
        ceil(buffer.shape[1], THREADSPERBLOCK[1]))
    )
    
    perform[(blockspergrid, THREADSPERBLOCK)](buffer)
    cuda.synchronize()



from numba.cuda.cudadrv.devicearray import DeviceNDArray
from .utils import grid_config, ceil
from numba import cuda

THREADSPERBLOCK = (64, 16)

@cuda.jit
def perform(buffer: DeviceNDArray, exps: DeviceNDArray, sum_exps: DeviceNDArray):
    batch_id, y_col = cuda.grid(2)
    
    if batch_id >= buffer.shape[0] or \
        y_col >= buffer.shape[2]:
            return None
    
    exp_value = buffer[batch_id, 0, y_col]
    exp_sum_value = sum_exps[batch_id, 0]
    
    res_value = (exp_value * exp_sum_value) - (exp_value * exp_value)
    res_value /= (exp_sum_value * exp_sum_value)
    
    buffer[batch_id, 0, y_col] = res_value

def softmax_derivate(buffer: DeviceNDArray, exps: DeviceNDArray, sum_exps: DeviceNDArray):
    blockspergrid = grid_config(
        (ceil(buffer.shape[0], THREADSPERBLOCK[0]),
         ceil(buffer.shape[2], THREADSPERBLOCK[1]))
    )
    
    perform[(blockspergrid, THREADSPERBLOCK)](buffer, exps, sum_exps)
    cuda.synchronize()



import numpy as np

def softmax_derivate(signals: np.ndarray, alphas: np.ndarray) -> np.ndarray:
    """evaluates: the derivate of softmax

    Args:
        signals (np.ndarray): [BATCH][1][N]
        alphas (np.ndarray): [BATCH][1][N]

    Returns:
        np.ndarray: [BATCH][1][N]
    """
    
    ret = []
    
    for signal, alpha in zip(signals, alphas):
        exp_signal = np.exp(signal)
        exp_signal_sum = exp_signal.sum()
        exp_times_sum = exp_signal * exp_signal_sum
        square_exp = exp_signal * exp_signal
        derivate = (exp_times_sum - square_exp)/(exp_signal_sum * exp_signal_sum)
        
        ret.append(derivate * alpha)
    
    return np.array(ret)



from numba.cuda.cudadrv.devicearray import DeviceNDArray
from .kernel import sigmoid_0_1_derivate, list_multiplication

def sigmoid2_derivate(signals: DeviceNDArray, alphas: DeviceNDArray, buffer: DeviceNDArray) -> DeviceNDArray:
    """evaluates sigmoid_0_1 derivate

    Args:
        signals (DeviceNDArray): [BATCH][1][N]
        alphas (DeviceNDArray): [BATCH][1][N]
        buffer (DeviceNDArray): [BATCH][1][N]

    Returns:
        DeviceNDArray: [BATCH][1][N]
    """
    sigmoid_0_1_derivate(buffer=buffer, signals=signals)
    list_multiplication(buffer=buffer, const_A=buffer, const_B=alphas)
    
    return buffer



import numpy as np
from numba import cuda
from Kinho.lib import cpu, gpu

EPS = 1e-8

def test_partial_gradient_gpu():
    DIM1 = 100
    DIM2 = 200
    
    weight_host = np.random.randn(DIM1, DIM2)
    gradient_host = np.random.randn(DIM1, DIM2)
    eta_host = np.random.randn(1)
    
    weight_dvc = cuda.to_device(weight_host)
    gradient_dvc = cuda.to_device(gradient_host)
    eta_dvc = cuda.to_device(eta_host)
    
    ret_cpu = cpu.partial_gradient(weight=weight_host, eta=eta_host, gradient=gradient_host)
    gpu.partial_gradient(weight=weight_dvc, eta=eta_dvc, gradient=gradient_dvc)
    ret_gpu = weight_dvc.copy_to_host()
    
    for i in range(DIM1):
        for j in range(DIM2):
            assert abs(ret_cpu[i, j] - ret_gpu[i, j]) < EPS



from numba.cuda.cudadrv.devicearray import DeviceNDArray
from numba import cuda
from .utils import ceil, grid_config

THREADSPERBLOCK = (64, 16)

@cuda.jit
def perform(buffer: DeviceNDArray, signals: DeviceNDArray, const_values: DeviceNDArray):
    batch_id, y_col = cuda.grid(2)
    
    if batch_id >= signals.shape[0] or \
        y_col >= signals.shape[2]:
            return None
    
    buffer[batch_id, 0, y_col] = signals[batch_id, 0, y_col] / const_values[batch_id, 0]

def divide_by_const(buffer: DeviceNDArray, signals: DeviceNDArray, const_values: DeviceNDArray):
    blockspergrid = grid_config(
        (ceil(signals.shape[0], THREADSPERBLOCK[0]),
        ceil(signals.shape[2], THREADSPERBLOCK[1]))
    )
    
    perform[(blockspergrid, THREADSPERBLOCK)](buffer, signals, const_values)
    cuda.synchronize()



from Kinho import Neural
from .shared import Shared

import random

shared = Shared()

LEARN_RATE = 0.4

def learn_rate(robot: Neural, epoch: int):
    images = shared.images()
    labels = shared.labels()
    
    zipped_data = [(img, lbl) for img, lbl in zip(images, labels)]
    random.shuffle(zipped_data)
    
    ALPHA = 0.1
    TEST_SIZE = int(len(zipped_data) * ALPHA)
    
    test = zipped_data[:TEST_SIZE]
    train = zipped_data[TEST_SIZE:]
    
    for i in range(epoch):
        for input in train:
            robot.learn(input[0], shared.densityArr(input[1], 10))
    
    hits = 0
    for input in test:
        out = robot.send(input[0])
        if shared.greaterIdx(out) == input[1]:
            hits += 1
    
    return hits/TEST_SIZE

def test_learn_rate_cpu():
    robot = Neural(sizes=[28*28, 15, 10], eta=1.0, gpu=False)
    
    assert learn_rate(robot=robot, epoch=1) >= LEARN_RATE

def test_learn_rate_gpu():
    robot = Neural(sizes=[28*28, 15, 10], eta=1.0, gpu=True)
    
    assert learn_rate(robot=robot, epoch=1) >= LEARN_RATE

def test_learn_rate_batch16_gpu():
    robot = Neural(sizes=[28*28, 15, 10], eta=1.0, gpu=True, mini_batch_size=16)
    
    assert learn_rate(robot=robot, epoch=10) >= 0.5



from Kinho.lib import cpu, gpu
import numpy as np
from numba import cuda

EPS = 1e-8

def test_softmax_derivate_gpu():
    BATCH = 256
    DIM = 100
    
    signals_host = np.random.randn(BATCH, 1, DIM)
    alphas_host = np.random.randn(BATCH, 1, DIM)
    extra_host = np.random.randn(BATCH, 1)
    
    signals_dvc = cuda.to_device(signals_host)
    alphas_dvc = cuda.to_device(alphas_host)
    extra_dvc = cuda.to_device(extra_host)
    buffer = cuda.device_array(shape=(BATCH, 1, DIM), dtype=np.float64)
    
    ret_cpu = cpu.softmax_derivate(signals=signals_host, alphas=alphas_host)
    gpu.softmax_derivate(signals=signals_dvc, alphas=alphas_dvc, extra=extra_dvc, buffer=buffer)
    ret_gpu = buffer.copy_to_host()
    
    for i in range(BATCH):
        for j in range(DIM):
            assert abs(ret_cpu[i, 0, j] - ret_gpu[i, 0, j]) < EPS



from numba.cuda.cudadrv.devicearray import DeviceNDArray
from .kernel import average_batch, memset_arr_2

def stochastic_gradient_descent(gradients: DeviceNDArray, buffer: DeviceNDArray) -> DeviceNDArray:
    """sgd for gradients

    Args:
        gradients (DeviceNDArray): [BATCH][N][M]
        buffer (DeviceNDArray): [N][M]

    Returns:
        DeviceNDArray: buffer
    """
    memset_arr_2(buffer=buffer)
    average_batch(buffer=buffer, gradients=gradients)
    
    return buffer



from numba.cuda.cudadrv.devicearray import DeviceNDArray
from .kernel import exponential, sum_list, memset_dim_1, list_multiplication
from .kernel import softmax_derivate as soft_derivate

def softmax_derivate(signals: DeviceNDArray, alphas: DeviceNDArray, extra: DeviceNDArray, buffer: DeviceNDArray) -> DeviceNDArray:
    exponential(buffer=buffer, signals=signals)
    memset_dim_1(buffer=extra)
    sum_list(buffer=extra, signals=buffer)
    soft_derivate(buffer=buffer, exps=buffer, sum_exps=extra)
    list_multiplication(buffer=buffer, const_A=buffer, const_B=alphas)
    
    return buffer



from numba.cuda.cudadrv.devicearray import DeviceNDArray
from numba import cuda, float64
from .utils import ceil, grid_config

THREADSPERBLOCK = (16, 64)

@cuda.jit
def perform(buffer: DeviceNDArray):
    row, col = cuda.grid(2)
    
    if row >= buffer.shape[0] or \
        col >= buffer.shape[1]:
            return None
    
    buffer[row, col] = float64(0.0)

def memset_arr_2(buffer: DeviceNDArray):
    blockspergrid = grid_config(
        (ceil(buffer.shape[0], THREADSPERBLOCK[0]),
         ceil(buffer.shape[1], THREADSPERBLOCK[1]))
    )

    perform[(blockspergrid, THREADSPERBLOCK)](buffer)
    cuda.synchronize()



from numba.cuda.cudadrv.devicearray import DeviceNDArray
from numba import cuda
from .utils import ceil, grid_config

THREADSPERBLOCK = (8, 128)

@cuda.jit
def perform(buffer: DeviceNDArray, A: DeviceNDArray, B: DeviceNDArray):
    batch_id, y_col = cuda.grid(2)

    if batch_id >= buffer.shape[0] or \
        y_col >= buffer.shape[2]:
            return None
    
    buffer[batch_id, 0, y_col] = A[batch_id, 0, y_col] * B[batch_id, 0, y_col]

def list_multiplication(buffer: DeviceNDArray, const_A: DeviceNDArray, const_B: DeviceNDArray):
    """buffer[i] = const_A[i] * const_B[i]

    Args:
        buffer (DeviceNDArray): [BATCH][1][N]
        const_A (DeviceNDArray): [BATCH][1][N]
        const_B (DeviceNDArray): [BATCH][1][N]
    """
    blockspergrid = grid_config(
        (ceil(buffer.shape[0], THREADSPERBLOCK[0]),
         ceil(buffer.shape[2], THREADSPERBLOCK[1]))
    )
    
    perform[(blockspergrid, THREADSPERBLOCK)](buffer, const_A, const_B)
    cuda.synchronize()
    


from .ceil import ceil

MINIMUM_BLOCK_SIZE = 28

def grid_1(shape):
    if shape[0] >= MINIMUM_BLOCK_SIZE:
        return shape

    block_x = shape[0] + (MINIMUM_BLOCK_SIZE - shape[0])

    return (block_x, )

def grid_2(shape):
    blocks = shape[0] * shape[1]
    
    if blocks >= MINIMUM_BLOCK_SIZE:
        return shape
    
    block_x = shape[0] + ceil(MINIMUM_BLOCK_SIZE - blocks, shape[1])
    
    return (block_x, shape[1])

def grid_3(shape):
    blocks = shape[0] * shape[1] * shape[2]
    
    if blocks >= MINIMUM_BLOCK_SIZE:
        return shape
    
    block_x = shape[0] + ceil(MINIMUM_BLOCK_SIZE - blocks, shape[1] * shape[2])
    
    return (block_x, shape[1], shape[2])

def grid_config(shape):
    if len(shape) == 1:
        return grid_1(shape)
    elif len(shape) == 2:
        return grid_2(shape)
    else:
        return grid_3(shape)



import numpy as np

def partial_gradient(weight: np.ndarray, eta: np.ndarray, gradient: np.ndarray) -> np.ndarray:
    """update: weight = weight - eta * gradient

    Args:
        weigth (np.ndarray): [N][M]
        eta (np.ndarray): [1]
        gradient (np.ndarray): [N][M]

    Returns:
        np.ndarray: weight
    """
    
    for i in range(weight.shape[0]):
        for j in range(weight.shape[1]):
            weight[i, j] = weight[i, j] - eta[0] * gradient[i, j]
    
    return weight



import numpy as np
import math
from timeit import default_timer as timer
from colorama import Fore, init
from .transfer import loadTo
from .brain import Wrapper, Builder
from .lib import cpu, gpu
from numba.cuda.cudadrv.devicearray import DeviceNDArray
from numba import cuda
from typing import Union

DATASTREAM = Union[DeviceNDArray, np.ndarray]

init()
MINIMUMBLOCKSIZE = 28

EPS = 1e-8

def ceil(A, B):
    return (A + B - 1) // B

class Neural(object):

    def __init__(self, sizes=None, brain_path=None, eta=0.01, gpu=False, mini_batch_size=1):
        if not sizes and not brain_path:
            raise TypeError('Should set `sizes` or `brain_path` params')
        
        self.__eta = np.array([eta])
        self.__num_layers = None
        self.__architecture = None
        self.__weights = None
        self.__biases = None
        self.__gpuMode = gpu
        self.__mini_batch = mini_batch_size
        self.__residual = None
        self.__target = None
        self.__fill = 0
        self.__logs = {}
        self.__tmp = {}

        t = timer()
        self.__setBrain(Builder(brain_path) if brain_path else None, sizes)
        self.__setMiniBatchMemo()
        self.__reserveMemo()
        t = timer() - t
    
    def __setBrain(self, brain, sizes):
        if brain:
            self.__num_layers = brain.size()
            self.__architecture = brain.architecture()
            self.__biases = brain.biases()
            self.__weights = brain.weights()
        else:
            self.__num_layers = len(sizes)
            self.__architecture = sizes
            self.__biases = [np.random.uniform(0, -1, x).reshape((1,x)) for x in self.__architecture[1:]]
            self.__weights = [np.random.uniform(-2,2,x*y).reshape((x, y)) for x,y in zip(self.__architecture[:-1], self.__architecture[1:])]

        if not self.__gpuMode:
            return None

        self.__eta = cuda.to_device(self.__eta)
        cuda_biases = [cuda.to_device(bias) for bias in self.__biases]
        self.__biases = cuda_biases

        cuda_weights = [cuda.to_device(weight) for weight in self.__weights]
        self.__weights = cuda_weights

    def __setMiniBatchMemo(self):
        if 2**int(math.log2(self.__mini_batch)) != self.__mini_batch:
            raise TypeError("Mini-batch size is invalid. You can use {}".format(2**int(math.log2(self.__mini_batch))))
        
        layers = len(self.__weights)
        activations = layers + 1
        selectors = 1
        
        edges = layers + activations + selectors
        
        self.__residual = []
        
        self.__residual.append(np.empty((self.__mini_batch, 1, self.__architecture[0])))
        
        for i in range(1, len(self.__architecture)):
            shape = (self.__mini_batch, 1, self.__architecture[i])
            self.__residual.append(np.empty(shape)) # after layer
            self.__residual.append(np.empty(shape)) # after activation
        
        shape = (self.__mini_batch, 1, self.__architecture[-1])
        self.__residual.append(np.empty(shape)) # selector
        
        self.__target = np.empty((self.__mini_batch, self.__architecture[-1]))
        
        if not self.__gpuMode:
            return None
        
        raw_residual = []
        
        for res in self.__residual:
            raw_residual.append(
                cuda.to_device(res)
            )
        
        raw_target = cuda.to_device(self.__target)
        
        self.__target = raw_target
        self.__residual = raw_residual
    
    def __getReserve(self, kernel, pos):
        if self.__gpuMode == False:
            return None
        
        return self.__buffer[kernel][pos]
    
    def __addMiniBatchOnShape(self, shape):
        new_shape = (self.__mini_batch, ) + shape
        return new_shape
    
    def __reserveArr(self, module_name, shape, streams, addMiniBatch=True):
        new_shape = shape
        
        if addMiniBatch:
            new_shape = self.__addMiniBatchOnShape(shape)

        arr = cuda.device_array(new_shape, dtype=np.float64)

        stream = cuda.stream()
        self.__buffer[module_name].append(cuda.to_device(arr, stream=stream))
        streams.append(stream)

    def __reserveActivation(self, streams):
        for size in self.__architecture:
            self.__reserveArr('activation', (1, size), streams)
    
    def __reserveLayer(self, streams):
        for size in self.__architecture[1:]:
            self.__reserveArr('layer', (1, size), streams)
    
    def __reserveSelector(self, streams):
        self.__reserveArr('selector', (1, self.__architecture[-1]), streams)
    
    def __reserveDLoss(self, streams):
        self.__reserveArr('d_loss', (1, self.__architecture[-1]), streams)
    
    def __reserveDSelector(self, streams):
        self.__reserveArr('d_selector', (1, self.__architecture[-1]), streams)

    def __reserveDActivation(self, streams):
        LEN = len(self.__architecture)

        for i in range(LEN-1, -1, -1):
            self.__reserveArr('d_activation', (1, self.__architecture[i]), streams)
    
    def __reserveTranspose(self, streams):
        LEN = len(self.__architecture)

        for i in range(LEN-2, -1, -1):
            self.__reserveArr('transpose', (self.__architecture[i], self.__architecture[i+1]), streams)
    
    def __reserveDLayer(self, streams):
        LEN = len(self.__architecture)

        for i in range(LEN-2, -1, -1):
            self.__reserveArr('d_layer', (1, self.__architecture[i]), streams)

    def __reserveExtra(self, streams):
        self.__reserveArr('extra', (1, ), streams)

    def __reserveWeight(self, streams):
        for i in range(len(self.__weights)):
            self.__reserveArr('weight', self.__weights[-i-1].shape, streams, addMiniBatch=False)
    
    def __reserveBias(self, streams):
        for i in range(len(self.__biases)):
            self.__reserveArr('bias', self.__biases[-i-1].shape, streams, addMiniBatch=False)
    
    def __reserveCost(self, streams):
        self.__reserveArr('cost', (1,), streams, addMiniBatch=False)

    def __reserveMemo(self):
        if self.__gpuMode == False:
            return None

        self.__buffer = {
            'activation': [],
            'layer': [],
            'selector': [],
            'd_loss': [],
            'd_selector': [],
            'd_activation': [],
            'transpose': [],
            'd_layer': [],
            'extra': [],
            'weight': [],
            'bias': [],
            'cost': [],
        }

        streams = []
        self.__reserveActivation(streams)
        self.__reserveLayer(streams)
        self.__reserveSelector(streams)
        self.__reserveDLoss(streams)
        self.__reserveDSelector(streams)
        self.__reserveDActivation(streams)
        self.__reserveTranspose(streams)
        self.__reserveDLayer(streams)
        self.__reserveExtra(streams)
        self.__reserveWeight(streams)
        self.__reserveBias(streams)
        self.__reserveCost(streams)

        for stream in streams:
            stream.synchronize()

    def logs(self):
        return self.__logs

    def __logger(self, method, delta, dbg=False):
        delta = round(delta * 1000, 3)

        mode = '[GPU] ' if self.__gpuMode else '[CPU] '
        name = mode + method

        if name not in self.__logs.keys():
            self.__logs[name] = []
        self.__logs[name].append(delta)

        color = Fore.GREEN
        if delta >= 1.0:
            color = Fore.YELLOW
        if delta >= 5.0:
            color = Fore.RED
        
        if dbg:
            print(color + "{}: {}ms".format(method, delta))

    def __loss(self, 
               predict: DATASTREAM, 
               target: DATASTREAM, 
               buffer: DeviceNDArray = None
        ) -> DATASTREAM:
        
        if self.__gpuMode:
            return gpu.mse(predict=predict, target=target, buffer=buffer)
        else:
            return cpu.mse(predict=predict, target=target)

    def __d_loss(self, 
                 predicts: DATASTREAM, 
                 targets: DATASTREAM, 
                 buffer: DeviceNDArray = None
        ) -> DATASTREAM:
        
        if self.__gpuMode:
            return gpu.mse_derivate(predicts=predicts, targets=targets, buffer=buffer)
        else:
            return cpu.mse_derivate(predicts=predicts, targets=targets)

    def __selector(self, 
                   signals: DATASTREAM, 
                   buffer: DeviceNDArray = None
        ) -> DATASTREAM:
        
        if self.__gpuMode:
            return gpu.softmax(signals=signals, extra=self.__getReserve('extra', 0), buffer=buffer)
        else:
            return cpu.softmax(signals=signals)

    def __d_selector(self, 
                     signals: DATASTREAM, 
                     alphas: DATASTREAM, 
                     buffer: DeviceNDArray = None
        ) -> DATASTREAM:
        
        if self.__gpuMode:
            return gpu.softmax_derivate(signals=signals, alphas=alphas, extra=self.__getReserve('extra', 0), buffer=buffer)
        else:
            return cpu.softmax_derivate(signals=signals, alphas=alphas)

    def __activation(self, 
                     signals: DATASTREAM, 
                     buffer: DeviceNDArray =None
        ) -> DATASTREAM:

        if self.__gpuMode:
            return gpu.sigmoid2(signals=signals, buffer=buffer)
        else:
            return cpu.sigmoid2(signals=signals)

    def __d_activation(self, 
                       signals: DATASTREAM, 
                       alphas: DATASTREAM, 
                       buffer: DeviceNDArray = None
        ) -> DATASTREAM:
        
        if self.__gpuMode:
            return gpu.sigmoid2_derivate(signals=signals, alphas=alphas, buffer=buffer)
        else:
            return cpu.sigmoid2_derivate(signals=signals, alphas=alphas)
    
    def __layer(self, 
                signals: DATASTREAM, 
                weight: DATASTREAM, 
                biase: DATASTREAM, 
                buffer: DeviceNDArray = None
        ) -> DATASTREAM:

        if self.__gpuMode:
            return gpu.dot_matrix(signals=signals, weight=weight, bias=biase, buffer=buffer)
        else:
            return cpu.dot_matrix(signals=signals, weight=weight, bias=biase)

    def __d_layer(self, 
                  const_weight: DATASTREAM, 
                  alphas: DATASTREAM, 
                  buffer: DeviceNDArray = None
        ) -> DATASTREAM:
        
        if self.__gpuMode:
            return gpu.dot_matrix_derivate(const_matrix=const_weight, alphas=alphas, buffer=buffer)
        else:
            return cpu.dot_matrix_derivate(const_matrix=const_weight, alphas=alphas)
    
    def __updateWeight(self, 
                       weight: DATASTREAM, 
                       eta: DATASTREAM, 
                       gradient: DATASTREAM, 
        ) -> DATASTREAM:
        
        if self.__gpuMode:
            return gpu.partial_gradient(weight=weight, eta=eta, gradient=gradient)
        else:
            return cpu.partial_gradient(weight=weight, eta=eta, gradient=gradient)
    
    def __transpose(self, 
                    signals: DATASTREAM, 
                    alphas: DATASTREAM, 
                    buffer: DeviceNDArray = None
        ) -> DATASTREAM:
        
        if self.__gpuMode:
            return gpu.transpose(signals=signals, alphas=alphas, buffer=buffer)
        else:
            return cpu.transpose(signals=signals, alphas=alphas)
    
    def __stochastic_gradient_descent(self, 
                                      gradients: DATASTREAM, 
                                      buffer: DeviceNDArray = None
        ) -> DATASTREAM:
        
        if self.__gpuMode:
            return gpu.stochastic_gradient_descent(gradients=gradients, buffer=buffer)
        else:
            return cpu.stochastic_gradient_descent(gradients=gradients)

    def __feedForward(self, x):
        t = timer()
        
        arr = self.__activation(x, buffer=self.__getReserve('activation', 0))

        activation_pointer = 1
        layer_pointer = 0

        for w, b in zip(self.__weights,self.__biases):
            x = self.__layer(arr, w, b, buffer=self.__getReserve('layer', layer_pointer))

            arr = self.__activation(x, buffer=self.__getReserve('activation', activation_pointer))

            activation_pointer += 1
            layer_pointer += 1

        y = self.__selector(x, buffer=self.__getReserve('selector', 0))

        t = timer() - t
        self.__logger("feedForward", t)
        return y

    def __backPropagation(self):
        t = timer()

        # feedForward
        x = self.__residual[0]

        activation_pointer = 1
        layer_pointer = 0
        residual_pointer = 1

        for w, b in zip(self.__weights, self.__biases):
            x = self.__layer(x, w, b, buffer=self.__getReserve('layer', layer_pointer))
            layer_pointer += 1

            self.__residual[residual_pointer] = x
            residual_pointer += 1

            x = self.__activation(x, buffer=self.__getReserve('activation', activation_pointer))
            activation_pointer += 1

            self.__residual[residual_pointer] = x
            residual_pointer += 1

        y = self.__selector(x, buffer=self.__getReserve('selector', 0))

        self.__residual[residual_pointer] = y
        residual_pointer += 1

        # backpropagation
        residual_pointer -= 1
        derror = self.__d_loss(self.__residual[residual_pointer], self.__target, buffer=self.__getReserve('d_loss', 0))
        
        residual_pointer -= 1
        derror = self.__d_selector(self.__residual[residual_pointer], derror, buffer=self.__getReserve('d_selector', 0))

        d_activation_pointer = 0
        d_layer_pointer = 0
        transpose_pointer = 0
        update_pointer = 0

        for l in range(1, self.__num_layers):
            w = self.__weights[-l]
            b = self.__biases[-l]

            residual_pointer -= 1
            derror = self.__d_activation(self.__residual[residual_pointer], derror, buffer=self.__getReserve('d_activation', d_activation_pointer))

            d_activation_pointer += 1
            
            residual_pointer -= 1
            nabla_w = self.__transpose(self.__residual[residual_pointer], derror, buffer=self.__getReserve('transpose', transpose_pointer))
            transpose_pointer += 1

            nabla_b = derror # error for each bias

            derror =  self.__d_layer(w, derror, buffer=self.__getReserve('d_layer', d_layer_pointer))

            d_layer_pointer += 1

            self.__weights[-l] = self.__updateWeight(
                self.__weights[-l], 
                self.__eta,
                self.__stochastic_gradient_descent(nabla_w, buffer=self.__getReserve('weight', update_pointer)))

            self.__biases[-l] = self.__updateWeight(
                self.__biases[-l], 
                self.__eta, 
                self.__stochastic_gradient_descent(nabla_b, buffer=self.__getReserve('bias', update_pointer)))
            
            update_pointer += 1

        t = timer() - t
        prefix = '[GPU] ' if self.__gpuMode else '[CPU] '
        self.__logger(prefix + "backpropagation", t, dbg=False)
    
    def __buildMsg(self, x):
        if self.__gpuMode:
            return cuda.to_device(x)
        return x
    
    def __insertIntoBatch(self, x, y):
        if 'img' not in self.__tmp.keys():
            self.__tmp['img'] = []
        
        if 'target' not in self.__tmp.keys():
            self.__tmp['target'] = []
        
        self.__tmp['img'].append(x)
        self.__tmp['target'].append(y)

        self.__fill += 1
        
        if self.__fill < self.__mini_batch:
            return None
        
        img = np.array(self.__tmp['img'])
        target = np.array(self.__tmp['target'])
        
        img = self.__buildMsg(img)
        target = self.__buildMsg(target)
        
        img = self.__activation(img, buffer=self.__getReserve('activation', 0))
        
        self.__residual[0] = img
        self.__target = target
        
        self.__backPropagation()
        self.__fill = 0
        
        self.__tmp['img'] = []
        self.__tmp['target'] = []

    def send(self, input):
        l = input
        
        t = timer()

        x = self.__buildMsg(np.array([[l]]))
        arr = self.__feedForward(x)

        hst, = loadTo(arr, mode='CPU')
        y = hst[0][0]

        t = timer() - t
        self.__logger("send", t)

        return y

    def learn(self, input, output):
        x, y = input, output

        t = timer()

        x = np.array([x])
        y = np.array([y])
        
        self.__insertIntoBatch(x, y)

        t = timer() - t
        self.__logger("learn", t)

    def cost(self, input, output):
        x, y = input, output
        
        t = timer()

        np_x = self.__buildMsg(np.array([[x]]))
        np_y = self.__buildMsg(np.array([[y]]))

        ret = self.__loss(self.__feedForward(np_x), np_y, buffer=self.__getReserve('cost', 0))

        t = timer() - t
        self.__logger("cost", t)
        return ret[0]
    
    def export(self, filename, path):
        layers = []

        for w, b in zip(self.__weights, self.__biases):
            w_host, b_host = loadTo(w, b, mode='CPU')
            layers.append((w_host, b_host))
        
        wrapper = Wrapper(layers)

        with open(path + filename + '.brain', 'wb') as file:
            file.write(wrapper.data())



import numpy as np

def transpose(signals: np.ndarray, alphas: np.ndarray) -> np.ndarray:
    """evaluates the gradient for a matrix_multiplication
    using signals as input and alphas as error

    Args:
        signals (np.ndarray): [BATCH][1][N]
        alphas (np.ndarray): [BATCH][1][M]

    Returns:
        np.ndarray: [BATCH][N][M]
    """
    
    ret = []
    
    for signal, alpha in zip(signals, alphas):
        ret.append(
            signal.transpose().dot(alpha)
        )
    
    return np.array(ret)


from Kinho.lib import cpu, gpu
from numba import cuda
import numpy as np

EPS = 1e-8

def test_dot_matrix_gpu():
    BATCH = 64
    N = 100
    M = 200

    signals_host = np.random.randn(BATCH, 1, N)
    weight_host = np.random.randn(N, M)
    bias_host = np.random.randn(1, M)
    
    signals_device = cuda.to_device(signals_host)
    weight_device = cuda.to_device(weight_host)
    bias_device = cuda.to_device(bias_host)
    buffer_device = cuda.device_array(shape=(BATCH, 1, M), dtype=np.float64)
    
    ret_cpu = cpu.dot_matrix(signals=signals_host, weight=weight_host, bias=bias_host)
    gpu.dot_matrix(signals=signals_device, weight=weight_device, bias=bias_device, buffer=buffer_device)
    ret_gpu = buffer_device.copy_to_host()
    
    for i in range(BATCH):
        for j in range(M):
            assert abs(ret_cpu[i, 0, j] - ret_gpu[i, 0, j]) < EPS



import numpy as np

def sigmoid2_derivate(signals: np.ndarray, alphas: np.ndarray) -> np.ndarray:
    """evaluates the derivate of sigmoid2:
    
    f'(x) = 2*exp(-x)/(1 + exp(-x))^2

    Args:
        signals (np.ndarray): [BATCH][1][N]
        alphas (np.ndarray): [BATCH][1][N]

    Returns:
        np.ndarray: [BATCH][1][N]
    """
    
    ret = []
    
    for signal, alpha in zip(signals, alphas):
        exp_signal = np.exp(-signal)
        exp_plus_one_square = (exp_signal + 1) * (exp_signal + 1)
        derivate = (2.0 * exp_signal) / exp_plus_one_square
        
        ret.append(
            derivate * alpha
        )
    
    return np.array(ret)



import numpy as np

def softmax(signals: np.ndarray) -> np.ndarray:
    """evaluates e^x/sum{e^x}

    Args:
        signals (np.ndarray): the signal sent

    Returns:
        np.ndarray: the evaluated signal
    """
    ret = []
    
    for signal in signals:
        z = np.exp(signal)
        sumT = z.sum()
        z = z/sumT
        ret.append(z)
    
    return np.array(ret)



import numpy as np

def stochastic_gradient_descent(gradients: np.ndarray) -> np.ndarray:
    """sgd for gradients

    Args:
        gradients (np.ndarray): [BATCH][N][M]

    Returns:
        np.ndarray: [N][M]
    """
    
    BATCH = gradients.shape[0]
    N = gradients.shape[1]
    M = gradients.shape[2]
    
    ret = np.empty(shape=(N, M), dtype=np.float64)
    
    for i in range(N):
        for j in range(M):
            acm = 0.0
            
            for k in range(BATCH):
                acm += gradients[k, i, j]
            
            ret[i, j] = acm
    
    return ret



from Kinho import Neural
from .shared import Shared

shared = Shared()

EPS = 1e-8

def send(robot: Neural):
    images = shared.images()
    
    ret = robot.send(input=images[0])
    
    assert len(ret) == 10
    
    acm = 0.0
    for i in range(10):
        acm += ret[i]
    
    assert abs(acm - 1.0) < EPS

def test_send_cpu():
    robot = Neural(sizes=[28*28, 15, 10], eta=0.1, gpu=False)
    
    send(robot)

def test_send_gpu():
    robot = Neural(sizes=[28*28, 15, 10], eta=0.1, gpu=True)

    send(robot)



import numpy as np

def dot_matrix_derivate(const_matrix: np.ndarray, alphas: np.ndarray) -> np.ndarray:
    """evaluates: alpha.dot(const_matrix.transpose())

    Args:
        const_matrix (np.ndarray): [N][M]
        alphas (np.ndarray): [BATCH][1][M]

    Returns:
        np.ndarray: [BATCH][1][N]
    """
    
    ret = []
    
    for alpha in alphas:
        ret.append(
            alpha.dot(const_matrix.transpose())
        )
    
    return np.array(ret)



from numba.cuda.cudadrv.devicearray import DeviceNDArray
from numba import cuda, float64
from .utils import grid_config, ceil

# (batch, A.y, B.y)
THREADSPERBLOCK = (1, 32, 32)

@cuda.jit
def perform(buffer: DeviceNDArray, const_a: DeviceNDArray, const_b: DeviceNDArray):
    batch_id, y_row, z_col = cuda.grid(3)

    tmp = cuda.shared.array(shape=THREADSPERBLOCK, dtype=float64)
    sA = cuda.shared.array(shape=(THREADSPERBLOCK[0], THREADSPERBLOCK[1]), dtype=float64)
    sB = cuda.shared.array(shape=(THREADSPERBLOCK[1], THREADSPERBLOCK[2]), dtype=float64)

    tx = cuda.threadIdx.x
    ty = cuda.threadIdx.y
    tz = cuda.threadIdx.z
    
    tmp[tx, ty, tz] = float64(0.0)
    
    if batch_id >= const_a.shape[0] or \
        y_row >= const_b.shape[0] or \
            z_col >= const_b.shape[1]:
                return None
    if tz == 0:
        sA[tx, ty] = const_a[batch_id, 0, y_row]
    
    if tx == 0:
        sB[ty, tz] = const_b[y_row, z_col]
    
    cuda.syncthreads()
    
    tmp[tx, ty, tz] = sA[tx, ty] * sB[ty, tz]
    
    cuda.syncthreads()
    
    if ty != 0:
        return None
    
    for i in range(1, cuda.blockDim.y):
        tmp[tx, 0, tz] += tmp[tx, i, tz]
    
    cuda.atomic.add(buffer, (batch_id, 0, z_col), tmp[tx, 0, tz])

def matrix_multiplication(buffer: DeviceNDArray, const_a: DeviceNDArray, const_b: DeviceNDArray) -> None:
    """ Perform buffer[batch_id] = const_a[batch_id] * const_b

    Args:
        buffer (DeviceNDArray): [batch][1][M]
        const_a (DeviceNDArray): [batch][1][N]
        const_b (DeviceNDArray): [N][M]
    """
    blockspergrid = grid_config(
        (ceil(buffer.shape[0], THREADSPERBLOCK[0]),
         ceil(const_b.shape[0], THREADSPERBLOCK[1]),
         ceil(const_b.shape[1], THREADSPERBLOCK[2]))
    )
    
    perform[(blockspergrid, THREADSPERBLOCK)](buffer, const_a, const_b)
    cuda.synchronize()



import numpy as np
from numba import cuda
from Kinho.lib import cpu, gpu

EPS = 1e-8

def test_mse_gpu():
    DIM = 100
    
    predict_host = np.random.randn(1, 1, DIM)
    target_host = np.random.randn(1, 1, DIM)
    
    predict_dvc = cuda.to_device(predict_host)
    target_dvc = cuda.to_device(target_host)
    buffer = cuda.device_array(shape=(1,), dtype=np.float64)
    
    ret_cpu = cpu.mse(predict=predict_host, target=target_host)
    gpu.mse(predict=predict_dvc, target=target_dvc, buffer=buffer)
    ret_gpu = buffer.copy_to_host()
    
    assert abs(ret_cpu[0] - ret_gpu[0]) < EPS



from Kinho.lib import cpu, gpu
import numpy as np
from numba import cuda

EPS = 1e-8

def test_softmax_gpu():
    BATCH = 128
    DIM = 40
    
    signals_host = np.random.randn(BATCH, 1, DIM)
    extra = np.empty(shape=(BATCH, 1))
    buffer = np.empty(shape=(BATCH, 1, DIM))
    
    signals_dvc = cuda.to_device(signals_host)
    extra_dvc = cuda.to_device(extra)
    buffer_dvc = cuda.to_device(buffer)
    
    ret_cpu = cpu.softmax(signals_host)
    gpu.softmax(signals=signals_dvc, extra=extra_dvc, buffer=buffer_dvc)
    ret_gpu = buffer_dvc.copy_to_host()
    
    for i in range(BATCH):
        for j in range(DIM):
            assert abs(ret_cpu[i, 0, j] - ret_gpu[i, 0, j]) < EPS



from numba.cuda.cudadrv.devicearray import DeviceNDArray
from numba import cuda, float64
from .utils import ceil, grid_config

THREADSPERBLOCKS = (1, 32, 32)

@cuda.jit
def perform(buffer: DeviceNDArray, const_A: DeviceNDArray, const_B: DeviceNDArray):
    batch_id, y_row, z_col = cuda.grid(3)
    
    sA = cuda.shared.array(shape=(THREADSPERBLOCKS[0], THREADSPERBLOCKS[1]), dtype=float64)
    sB = cuda.shared.array(shape=(THREADSPERBLOCKS[0], THREADSPERBLOCKS[2]), dtype=float64)
    
    tx = cuda.threadIdx.x
    ty = cuda.threadIdx.y
    tz = cuda.threadIdx.z
    
    if batch_id >= buffer.shape[0] or \
        y_row >= buffer.shape[1] or \
            z_col >= buffer.shape[2]:
                return None
    
    if ty == 0:
        sB[tx, tz] = const_B[batch_id, 0, z_col]
    
    if tz == 0:
        sA[tx, ty] = const_A[batch_id, 0, y_row]
    
    cuda.syncthreads()

    buffer[batch_id, y_row, z_col] = sA[tx, ty] * sB[tx, tz]

def transpose_dot(buffer: DeviceNDArray, const_A: DeviceNDArray, const_B: DeviceNDArray):
    blockspergrid = grid_config(
        (ceil(buffer.shape[0], THREADSPERBLOCKS[0]),
         ceil(buffer.shape[1], THREADSPERBLOCKS[1]),
         ceil(buffer.shape[2], THREADSPERBLOCKS[2]))
    )
    
    perform[(blockspergrid, THREADSPERBLOCKS)](buffer, const_A, const_B)
    cuda.synchronize()



from numba.cuda.cudadrv.devicearray import DeviceNDArray
from .kernel import sigmoid_0_1

def sigmoid2(signals: DeviceNDArray, buffer: DeviceNDArray) -> DeviceNDArray:
    """evaluates the sigmoid in (-1, 1)

    Args:
        signals (DeviceNDArray): [batch][1][N]
        buffer (DeviceNDArray): [batch][1][N]

    Returns:
        (DeviceNDArray): [batch][1][N]
    """
    sigmoid_0_1(buffer=buffer, signals=signals)
    
    return buffer



import numpy as np

def dot_matrix(
    signals: np.ndarray, weight: np.ndarray, bias: np.ndarray
) -> np.ndarray:
    """
    Args:
        signals (np.array): the batch input for network -> shape(minibatch, 1, dim)
        weight (np.array): the weights of a single layer -> shape(dim, dim2)
        bias (np.array): the biase of a single layer -> (1, dim2)

    Returns:
        np.array: the output batch for layer -> shape(minibatch, 1, dim2)
    """
    ret = []

    for signal in signals:
        ret.append(signal.dot(weight) + bias)

    return np.array(ret)



import numpy as np
from numba import cuda
from Kinho.lib import cpu, gpu

EPS = 1e-8

def test_mse_derivate_gpu():
    BATCH = 256
    DIM = 40
    
    predict_host = np.random.randn(BATCH, 1, DIM)
    target_host = np.random.randn(BATCH, 1, DIM)
    buffer_host = np.random.randn(BATCH, 1, DIM)
    
    predict_dvc = cuda.to_device(predict_host)
    target_dvc = cuda.to_device(target_host)
    buffer_dvc = cuda.to_device(buffer_host)
    
    ret_cpu = cpu.mse_derivate(predict_host, target_host)
    gpu.mse_derivate(predict_dvc, target_dvc, buffer_dvc)
    ret_gpu = buffer_dvc.copy_to_host()
    
    for i in range(BATCH):
        for j in range(DIM):
            assert (ret_cpu[i, 0, j] - ret_gpu[i, 0, j]) <= EPS



from Kinho import Neural
from .shared import Shared
import random

shared = Shared()

def mirror(tmp: Neural):
    images = shared.images()
    labels = shared.labels()
    
    zipped_data = [(img, lbl) for img, lbl in zip(images, labels)]
    random.shuffle(zipped_data)
    
    PARTIAL_TRAIN = int(len(zipped_data) * 0.1)
    
    for input in zipped_data[:PARTIAL_TRAIN]:
        tmp.learn(input[0], shared.densityArr(input[1], 10))
    
    tmp.export(filename='tmp_test', path='./tmp/')
    
    robot_cpu = Neural(brain_path='./tmp/tmp_test.brain', eta=0.01, gpu=False)
    robot_gpu = Neural(brain_path='./tmp/tmp_test.brain', eta=0.01, gpu=True)
    
    random.shuffle(zipped_data)
    
    PARTIAL = 100
    EPS = 1e-8
    
    for i in range(PARTIAL):
        cpu_pred = robot_cpu.send(zipped_data[i][0])
        gpu_pred = robot_gpu.send(zipped_data[i][0])
        
        for c_pred, g_pred in zip(cpu_pred, gpu_pred):
            assert abs(c_pred - g_pred) < EPS

def test_mirror_learnWithGPU_gpu():
    tmp = Neural([28*28, 15, 10], eta=0.01, gpu=True)
    mirror(tmp)

def test_mirror_learnWithCPU_gpu():
    tmp = Neural([28*28, 15, 10], eta=0.01, gpu=False)
    mirror(tmp)



from numba.cuda.cudadrv.devicearray import DeviceNDArray
from .kernel import transpose_dot

def transpose(signals: DeviceNDArray, alphas: DeviceNDArray, buffer: DeviceNDArray) -> DeviceNDArray:
    """evaluates the gradient for a matrix_multiplication
    using signals as input and alphas as error

    Args:
        signals (DeviceNDArray): [BATCH][1][N]
        alphas (DeviceNDArray): [BATCH][1][M]
        buffer (DeviceNDArray): [BATCH][N][M]

    Returns:
        DeviceNDArray: buffer
    """
    transpose_dot(buffer=buffer, const_A=signals, const_B=alphas)
    
    return buffer



import numpy as np

def mse(predict: np.ndarray, target: np.ndarray) -> np.ndarray:
    """mse error for predict and target

    Args:
        predict (np.ndarray): [BATCH][1][N]
        target (np.ndarray): [BATCH][1][N]

    Returns:
        np.ndarray: [1]
    """
    
    diff = predict - target
    diff_square = diff * diff
    ret = diff_square.sum() / target.shape[2]
    
    return np.array([ret])



from numba.cuda.cudadrv.devicearray import DeviceNDArray
from .kernel import exponential, sum_list, divide_by_const, memset_dim_1

def softmax(signals: DeviceNDArray, extra: DeviceNDArray, buffer: DeviceNDArray):
    """e^(z)/sum{e^z}

    Args:
        signals (DeviceNDArray): [batch][1][N]
        extra (DeviceNDArray): [batch][1]
        buffer (DeviceNDArray): [batch][1][N]
    
    returns:
        buffer
    """
    exponential(buffer=buffer, signals=signals)
    memset_dim_1(buffer=extra)
    sum_list(buffer=extra, signals=buffer)
    divide_by_const(buffer=buffer, signals=buffer, const_values=extra)
    
    return buffer



from numba.cuda.cudadrv.devicearray import DeviceNDArray
from .kernel import inverse_matrix_multiplication, memset_arr_1

def dot_matrix_derivate(const_matrix: DeviceNDArray, alphas: DeviceNDArray, buffer: DeviceNDArray) -> DeviceNDArray:
    """alpha.dot(const_matrix.transpose())

    Args:
        const_matrix (DeviceNDArray): [N][M]
        alphas (DeviceNDArray): [BATCH][1][M]
        buffer (DeviceNDArray): [BATCH][1][N]
    """
    
    memset_arr_1(buffer)
    inverse_matrix_multiplication(buffer=buffer, const_matrix=const_matrix, alphas=alphas)
    
    return buffer



from numba.cuda.cudadrv.devicearray import DeviceNDArray
from .kernel import partial_gradient as partial

def partial_gradient(weight: DeviceNDArray, eta: DeviceNDArray, gradient: DeviceNDArray) -> DeviceNDArray:
    """update: weight - eta * gradient

    Args:
        weight (DeviceNDArray): [N][M]
        eta (DeviceNDArray): [1]
        gradient (DeviceNDArray): [N][M]

    Returns:
        DeviceNDArray: [N][M]
    """
    
    partial(buffer=weight, const_A=weight, const_B=gradient, eta=eta)
    
    return weight



from numba.cuda.cudadrv.devicearray import DeviceNDArray
from numba import cuda
from .utils import ceil, grid_config

THREADSPERBLOCK = (128, 8)

@cuda.jit
def perform(buffer: DeviceNDArray, predicts: DeviceNDArray, targets: DeviceNDArray):
    batch_id, y_col = cuda.grid(2)
    
    if batch_id >= buffer.shape[0] or \
        y_col >= buffer.shape[2]:
            return None
    
    N = buffer.shape[2]
    
    buffer[batch_id, 0, y_col] = (2.0 / N) * (predicts[batch_id, 0, y_col] - targets[batch_id, 0, y_col])

def mse_derivate(buffer: DeviceNDArray, predicts: DeviceNDArray, targets: DeviceNDArray):
    blockspergrid = grid_config(
        (ceil(buffer.shape[0], THREADSPERBLOCK[0]),
         ceil(buffer.shape[2], THREADSPERBLOCK[1]))
    )
    
    perform[(blockspergrid, THREADSPERBLOCK)](buffer, predicts, targets)
    cuda.synchronize()



from .dot_matrix import dot_matrix
from .sigmoid2 import sigmoid2
from .softmax import softmax
from .mse_derivate import mse_derivate
from .softmax_derivate import softmax_derivate
from .sigmoid2_derivate import sigmoid2_derivate
from .transpose import transpose
from .dot_matrix_derivate import dot_matrix_derivate
from .stochastic_gradient_descent import stochastic_gradient_descent
from .partial_gradient import partial_gradient
from .mse import mse



from numba.cuda.cudadrv.devicearray import DeviceNDArray
from numba import cuda, float64
from .utils import grid_config, ceil

THREADSPERBLOCK = (32, 32)

@cuda.jit
def perform(buffer: DeviceNDArray):
    batch_id, y_col = cuda.grid(2)
    
    if batch_id >= buffer.shape[0] or \
        y_col >= buffer.shape[2]:
            return None
    
    buffer[batch_id, 0, y_col] = float64(0.0)

def memset_arr_1(buffer: DeviceNDArray):
    """reset arr with 0s

    Args:
        buffer (DeviceNDArray): [BATCH][1][N]
    """
    blockspergrid = grid_config(
        (ceil(buffer.shape[0], THREADSPERBLOCK[0]),
         ceil(buffer.shape[2], THREADSPERBLOCK[1]))
    )
    
    perform[(blockspergrid, THREADSPERBLOCK)](buffer)
    cuda.synchronize()



from Kinho import Neural
from .shared import Shared

shared = Shared()

def cost(robot: Neural):
    images = shared.images()
    labels = shared.labels()
    
    error = robot.cost(images[0], shared.densityArr(labels[0], 10))
    
    assert isinstance(error, float)

def test_cost_cpu():
    robot = Neural(sizes=[28*28, 15, 10], eta=0.1, gpu=False)
    
    cost(robot)

def test_cost_gpu():
    robot = Neural(sizes=[28*28, 15, 10], eta=0.1, gpu=True)

    cost(robot)



from numba.cuda.cudadrv.devicearray import DeviceNDArray
from numba import cuda, float64
from .utils import ceil, grid_config

THREADSPERBLOCK = (32, 32)

@cuda.jit
def perform(buffer: DeviceNDArray, const_host: DeviceNDArray):
    batch_id, y_col = cuda.grid(2)
    
    sH = cuda.shared.array(shape=(THREADSPERBLOCK[1],), dtype=float64)
    
    tx = cuda.threadIdx.x
    ty = cuda.threadIdx.y
    
    if tx == 0:
        sH[ty] = const_host[0, y_col]
    
    if batch_id >= buffer.shape[0] or \
        y_col >= const_host.shape[1]:
            return None
    
    cuda.syncthreads()
    
    buffer[batch_id, 0, y_col] = sH[ty]

def copy(buffer: DeviceNDArray, const_host: DeviceNDArray):
    """Perform: buffer[batch_id] = const_host

    Args:
        buffer (DeviceNDArray): [batch][1][N]
        const_host (DeviceNDArray): [1][N]
    """

    blockspergrid = grid_config(
        (ceil(buffer.shape[0], THREADSPERBLOCK[0]), ceil(const_host.shape[1], THREADSPERBLOCK[1]))
    )
    
    perform[(blockspergrid, THREADSPERBLOCK)](buffer, const_host)
    cuda.synchronize()



from numba.cuda.cudadrv.devicearray import DeviceNDArray
from numba import cuda, float64
from .utils import ceil, grid_config

THREADSPERBLOCK = (1024,)

@cuda.jit
def perform(buffer):
    x = cuda.grid(1)
    
    if x >= buffer.shape[0]:
        return None
    
    buffer[x] = float64(0.0)

def memset_constant(buffer: DeviceNDArray):
    blockspergrid = grid_config(
        (ceil(1, THREADSPERBLOCK[0]),)
    )
    
    perform[(blockspergrid, THREADSPERBLOCK)](buffer)
    cuda.synchronize()



from .copy import copy
from .matrix_multiplication import matrix_multiplication
from .sigmoid_0_1 import sigmoid_0_1
from .exponential import exponential
from .sum_list import sum_list
from .divide_by_const import divide_by_const
from .memset_dim_1 import memset_dim_1
from .mse_derivate import mse_derivate
from .softmax_derivate import softmax_derivate
from .list_multiplication import list_multiplication
from .sigmoid_0_1_derivate import sigmoid_0_1_derivate
from .transpose_dot import transpose_dot
from .inverse_matrix_multiplication import inverse_matrix_multiplication
from .memset_arr_1 import memset_arr_1
from .average_batch import average_batch
from .memset_arr_2 import memset_arr_2
from .partial_gradient import partial_gradient
from .mse import mse
from .memset_constant import memset_constant



from numba.cuda.cudadrv.devicearray import DeviceNDArray
from .utils import grid_config, ceil
from numba import cuda
import math

THREADSPERBLOCK = (8, 128)

@cuda.jit
def perform(buffer: DeviceNDArray, const_z: DeviceNDArray):
    batch_id, y_col = cuda.grid(2)
    
    if batch_id >= buffer.shape[0] or \
        y_col >= const_z.shape[2]:
            return None
    
    buffer[batch_id, 0, y_col] = 2.0 * (1.0 / (1.0 + math.exp(-const_z[batch_id, 0, y_col]))) - 1.0

def sigmoid_0_1(buffer: DeviceNDArray, signals: DeviceNDArray):
    """sigmoid with media: 0, std-dvt: 1

    Args:
        signals (DeviceNDArray): [batch][1][N]
        buffer (DeviceNDArray): [batch][1][N]
    """
    blockspergrid = grid_config(
        (ceil(signals.shape[0], THREADSPERBLOCK[0]), 
        ceil(signals.shape[2], THREADSPERBLOCK[1]))
    )
    
    perform[(blockspergrid, THREADSPERBLOCK)](buffer, signals)
    cuda.synchronize()



from numba.cuda.cudadrv.devicearray import DeviceNDArray
from .kernel import memset_constant, mse as kernel_mse

def mse(predict: DeviceNDArray, target: DeviceNDArray, buffer: DeviceNDArray) -> DeviceNDArray:
    """evaluates mse error

    Args:
        predict (DeviceNDArray): [BATCH][1][N]
        target (DeviceNDArray): [BATCH][1][N]
        buffer (DeviceNDArray): [1]

    Returns:
        DeviceNDArray: buffer
    """
    
    memset_constant(buffer)
    kernel_mse(buffer=buffer, predict=predict, target=target)
    
    return buffer



from numba.cuda.cudadrv.devicearray import DeviceNDArray
from numba import cuda, float64
from .utils import ceil, grid_config

THREADSPERBLOCK = (1, 32, 32)

@cuda.jit
def perform(buffer: DeviceNDArray, const_matrix: DeviceNDArray, alphas: DeviceNDArray):
    batch_id, y_row, z_col = cuda.grid(3)
    
    sMatrix = cuda.shared.array(shape=(THREADSPERBLOCK[1], THREADSPERBLOCK[2]), dtype=float64)
    sAlpha = cuda.shared.array(shape=(THREADSPERBLOCK[0], THREADSPERBLOCK[1]), dtype=float64)
    tmp = cuda.shared.array(shape=THREADSPERBLOCK, dtype=float64)
    
    tx = cuda.threadIdx.x
    ty = cuda.threadIdx.y
    tz = cuda.threadIdx.z
    
    tmp[tx, ty, tz] = float64(0.0)
    
    if batch_id >= buffer.shape[0] or \
        y_row >= const_matrix.shape[1] or \
            z_col >= const_matrix.shape[0]: # inverse
                return None

    if tx == 0:
        sMatrix[ty, tz] = const_matrix[z_col, y_row] # inverse
    
    if tz == 0:
        sAlpha[tx, ty] = alphas[batch_id, 0, y_row]
    
    cuda.syncthreads()
    
    tmp[tx, ty, tz] = sAlpha[tx, ty] * sMatrix[ty, tz]
    
    cuda.syncthreads()
    
    if ty != 0:
        return None
    
    for i in range(1, cuda.blockDim.y):
        tmp[tx, 0, tz] += tmp[tx, i, tz]
    
    cuda.atomic.add(buffer, (batch_id, 0, z_col), tmp[tx, 0, tz])

def inverse_matrix_multiplication(buffer: DeviceNDArray, const_matrix: DeviceNDArray, alphas: DeviceNDArray):
    blockspergrid = grid_config(
        (ceil(buffer.shape[0], THREADSPERBLOCK[0]),
         ceil(const_matrix.shape[1], THREADSPERBLOCK[1]), # inverse shape
         ceil(const_matrix.shape[0], THREADSPERBLOCK[2]))
    )
    
    perform[(blockspergrid, THREADSPERBLOCK)](buffer, const_matrix, alphas)
    cuda.synchronize()



from numba.cuda.cudadrv.devicearray import DeviceNDArray
from numba import cuda
from .utils import ceil, grid_config

THREADSPERBLOCK = (1024,)

@cuda.jit
def perform(buffer: DeviceNDArray, predict: DeviceNDArray, target: DeviceNDArray):
    x = cuda.grid(1)
    
    if x >= target.shape[2]:
        return None
    
    diff = predict[0, 0, x] - target[0, 0, x]
    diff_square_divide_N = (diff * diff)/target.shape[2]
    
    cuda.atomic.add(buffer, 0, diff_square_divide_N)

def mse(buffer: DeviceNDArray, predict: DeviceNDArray, target: DeviceNDArray):
    blockspergrid = grid_config(
        (ceil(target.shape[2], THREADSPERBLOCK[0]),)
    )
    
    perform[(blockspergrid, THREADSPERBLOCK)](buffer, predict, target)
    cuda.synchronize()



from numba.cuda.cudadrv.devicearray import DeviceNDArray
from numba import cuda, float64
from .utils import ceil, grid_config

THREADSPERBLOCK = (64, 1, 16)

@cuda.jit
def perform(buffer: DeviceNDArray, gradients: DeviceNDArray):
    batch_id, y_row, z_col = cuda.grid(3)
    
    tx = cuda.threadIdx.x
    ty = cuda.threadIdx.y
    tz = cuda.threadIdx.z
    
    tmp = cuda.shared.array(shape=THREADSPERBLOCK, dtype=float64)
    tmp[tx, ty, tz] = float64(0.0)
    
    if batch_id >= gradients.shape[0] or \
        y_row >= gradients.shape[1] or \
            z_col >= gradients.shape[2]:
                return None
    
    cuda.syncthreads()
    
    tmp[tx, ty, tz] = gradients[batch_id, y_row, z_col]
    
    cuda.syncthreads()
    
    if tx != 0:
        return None
    
    for k in range(1, cuda.blockDim.x):
        tmp[0, ty, tz] += tmp[k, ty, tz]
    
    cuda.atomic.add(buffer, (y_row, z_col), tmp[0, ty, tz])

def average_batch(buffer: DeviceNDArray, gradients: DeviceNDArray):
    blockspergrid = grid_config(
        (ceil(gradients.shape[0], THREADSPERBLOCK[0]),
         ceil(gradients.shape[1], THREADSPERBLOCK[1]),
         ceil(gradients.shape[2], THREADSPERBLOCK[2]))
    )
    
    perform[(blockspergrid, THREADSPERBLOCK)](buffer, gradients)
    cuda.synchronize()



from numba.cuda.cudadrv.devicearray import DeviceNDArray
from .kernel import matrix_multiplication, copy

def dot_matrix(
    signals: DeviceNDArray,
    weight: DeviceNDArray,
    bias: DeviceNDArray,
    buffer: DeviceNDArray,
):
    """buffer[batch_id] = signails[batch_id] * weight + bias

    Args:
        signals (DeviceNDArray): [batch_id][1][N]
        weight (DeviceNDArray): [N][M]
        bias (DeviceNDArray): [1][M]
        buffer (DeviceNDArray): [batch_id][1][M]

    Returns:
        buffer
    """
    copy(buffer, bias)
    matrix_multiplication(buffer=buffer, const_a=signals, const_b=weight)

    return buffer



import numpy as np

def mse_derivate(predicts: np.ndarray, targets: np.ndarray) -> np.ndarray:
    """evaluates: error[i] = (2/N) * (predict[i] - target[i])

    Args:
        predicts (np.ndarray): [BATCH][1][N]
        targets (np.ndarray): [BATCH][1][N]

    Returns:
        np.ndarray: [BATCH][1][N]
    """
    ret = []
    
    for predict, target in zip(predicts, targets):
        ret.append(
            2.0 * (predict - target) / predict.shape[1]
        )
    
    return np.array(ret)


from Kinho.lib import cpu, gpu
import numpy as np
from numba import cuda

EPS = 1e-8

def test_sigmoid2_gpu():
    BATCH = 256
    DIM = 1000
    
    signals_host = np.random.randn(BATCH, 1, DIM)
    
    signals_device = cuda.to_device(signals_host)
    buffer_device = cuda.device_array(shape=(BATCH, 1, DIM), dtype=np.float64)
    
    ret_cpu = cpu.sigmoid2(signals=signals_host)
    gpu.sigmoid2(signals=signals_device, buffer=buffer_device)
    ret_gpu = buffer_device.copy_to_host()
    
    for i in range(BATCH):
        for j in range(DIM):
            assert abs(ret_cpu[i, 0, j] - ret_gpu[i, 0, j]) < EPS



import numpy as np

def sigmoid2(signals: np.ndarray) -> np.ndarray:
    """evaluates : 1/(1 + e^(-z)) - 1.0
    all values are mapped into a range (-1.0, 1.0)

    Args:
        signals (np.ndarray): [batch][1][N]

    Returns:
        np.ndarray: [batch][1][N]
    """
    ret = []
    
    for signal in signals:
        ret.append(
            2.0 * (1.0 / (1.0 + np.exp(-signal))) - 1.0
        )
    
    return np.array(ret)
